"""
This module contains the RyskOptionMarket class.
"""
from dataclasses import dataclass
from datetime import datetime

from rysk_client.src.utils import get_contract


@dataclass
class RyskOptionMarket:
    """Rysk option market."""

    strike: float
    expiration: int
    is_put: bool

    @classmethod
    def from_series(cls, series):
        """Returns a RyskOptionMarket from a series"""
        return cls(
            strike=series["strike"],
            expiration=series["expiration"],
            is_put=series["isPut"],
        )

    def _parse_name(self):
        """Returns the name of the option market into the following format.

        NAME = "ETH-30JUN23-2200-C"
        """
        _type = "P" if self.is_put else "C"
        _expiration = datetime.fromtimestamp(self.expiration).strftime("%d%b%y").upper()
        _strike = int(self.strike / 10**18)
        return f"ETH-{_expiration}-{_strike}-{_type}"

    @property
    def name(self):
        """Returns the name of the option market"""
        return self._parse_name()

    def __str__(self):
        return f"RyskOptionMarket({self.name})"


class RyskOptionMarketManager:
    """Class to represent the RyskOptionMarketManager."""

    def fetch_option_markets(self):
        """Fetches the option markets from the API"""
        markets = []
        for expiration in self._fetch_expirations():
            put_strikes = self.option_catalogue.functions.getOptionDetails(
                expiration, True
            ).call()
            call_strikes = self.option_catalogue.functions.getOptionDetails(
                expiration, False
            ).call()

            for strike in put_strikes:
                markets.append(RyskOptionMarket(expiration, strike, True))

            for strike in call_strikes:
                markets.append(RyskOptionMarket(expiration, strike, False))
        self.option_markets = markets
        return markets

    def __init__(self, web3):
        self.web3 = web3
        self.option_markets = {}
        self.option_catalogue = get_contract("option_catalogue", web3)
        self.fetch_option_markets()

    def fetch_option_market(self, name):
        """
        Fetches a specific option market from the smart contract.
        """

    def _fetch_expirations(self):
        """Returns the expirations from the smart contract."""
        return self.option_catalogue.functions.getExpirations().call()



"""
Helper functions for the rysk_client package.
"""
import json
from copy import deepcopy

from web3 import Web3

from rysk_client.src.constants import ADDRESSES, DEFAULT_ENCODING, RPC_URL


def get_contract(name, web3):
    """Returns a web3 contract instance for the given contract name"""
    spec = ADDRESSES[name]
    with open(spec["path"], "r", encoding=DEFAULT_ENCODING) as abi:
        abi = json.loads(abi.read())["abi"]
    res = deepcopy(spec)
    del res["path"]
    res["abi"] = abi
    return web3.eth.contract(**res)


def get_web3() -> Web3:
    """Returns a web3 instance connected to RPC_URL"""
    web3 = Web3(Web3.HTTPProvider(RPC_URL))
    return web3



"""
Module to interact with the Rysk subgraph.
"""
import json
from dataclasses import dataclass

import requests

from rysk_client.src.constants import DEFAULT_TIMEOUT, SUBGRAPH_URL

MARKET_SUBGRAPH_QUERY = """
{series 
  {   
    id 
    expiration 
    netDHVExposure 
    strike
    isPut
    isBuyable
    isSellable
  }
}
"""


@dataclass
class SubgraphClient:
    """Simple client to interact with the Rysk subgraph."""

    url: str = SUBGRAPH_URL

    def _query(self, query):
        """Simple function to call a subgraph query."""
        headers = {"Content-Type": "application/json"}
        subgraph_query = {"query": query}
        response = requests.post(
            url=self.url,
            headers=headers,
            data=json.dumps(subgraph_query),
            timeout=DEFAULT_TIMEOUT,
        )
        data = json.loads(response.content)["data"]
        return data

    def query_markets(self):
        """Query the subgraph for markets."""
        return self._query(MARKET_SUBGRAPH_QUERY)["series"]



"""
Module to manage positions.
"""
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Optional

from rysk_client.src.collateral import Collateral
from rysk_client.src.rysk_option_market import RyskOptionMarket


class OrderSide(Enum):
    """Enum to represent the side of an order."""

    BUY = "buy"
    SELL = "sell"


class Order:
    """Class to represent an order"""

    price: float
    amount: float
    order_id: str
    order_side: OrderSide
    order_type: str = "market"


class PositionSide(Enum):
    """Enum to represent the side of a position."""

    LONG = "long"
    SHORT = "short"


@dataclass
class Position:
    """Class to represent a position."""

    option_market: RyskOptionMarket
    orders: List[Order]
    side: PositionSide
    size: float = 0
    pnl: float = 0
    collateral: Collateral = Collateral.USDC
    owner_address: Optional[str] = None


class PositionManager:
    """Class to manage positions"""

    positions: List[Position]
    collateral: Dict[str, float]

    def __init__(self):
        self.positions = []
        self.collateral = {}
        self.pnl = {}
        self.collateral_value = 0
        self.collateral_value_usd = 0

    def add_position(self, position: Position):
        """Adds a position to the position manager"""

    def update_position(self, position: Position):
        """Updates a position in the position manager"""

    def get_positions(self, address: Optional[str] = None) -> List[Position]:
        """Returns a list of positions"""
        if address:
            return [
                position
                for position in self.positions
                if position.owner_address == address
            ]
        return self.positions

    def _fetch_positions(self):
        """Fetches the positions from the API"""

    def _fetch_orders(self):
        """Fetches the orders from the API"""



"""
Web3 client for RyskFinance.
"""
from web3 import Web3, contract

from rysk_client.src.collateral import Collateral
from rysk_client.src.utils import get_contract, get_web3

w3 = Web3(Web3.HTTPProvider("https://arbitrum-goerli.rpc.thirdweb.com"))

addresses = {
    "beyond_pricer": {
        "path": "./contracts/BeyondPricer.sol/BeyondPricer.json",
        "address": "0xc939df369C0Fc240C975A6dEEEE77d87bCFaC259",
    },
    "opyn_controller": {
        "path": "contracts/packages/opyn/core/Controller.sol/Controller.json",
        "address": "0x11a602a5F5D823c103bb8b7184e22391Aae5F4C2",
    },
    "option_exchange": {
        "path": "contracts/OptionExchange.sol/OptionExchange.json",
        "address": "",
    },
}


def filter_by_human_format(data_frame, human_format):
    """
    Filter the DataFrame based on the human-readable format and return a single row.
    """

    filtered_option = data_frame[data_frame["human_strike"] == human_format]
    if len(filtered_option) > 0:
        return filtered_option.iloc[0]  # Return the first row if found
    return None  # Return None if no matching row is found


class Web3Client:
    """Client for the RyskFinance protocol."""

    beyond_pricer: contract.Contract = get_contract("beyond_pricer", get_web3())
    opyn_controller: contract.Contract = get_contract("opyn_controller", get_web3())

    def get_options_prices(
        self, option_data, amount=1000000000000000000, side="buy", collateral="weth"
    ):
        """,
        We call the beyond pricer to determine the prices for a market
        huge thanks to 0xPawel2 and Jib &&
        """
        if side not in ["buy", "sell"]:
            raise ValueError("Side must be buy or sell")

        if not Collateral.is_supported(collateral):
            raise TypeError(f"Collateral {collateral} is not supported")
        # here we call the contract functions

        option_series = (
            int(option_data["expiration"]),
            int(option_data["strike"]),
            bool(option_data["isPut"]),
            "0x3b3a1dE07439eeb04492Fa64A889eE25A130CDd3",
            "0x408c5755b5c7a0a28D851558eA3636CfC5b5b19d",
            "0x408c5755b5c7a0a28D851558eA3636CfC5b5b19d",
        )
        result = self.beyond_pricer.functions.quoteOptionPrice(
            option_series,
            int(amount),
            side == "sell",
            int(option_data["netDHVExposure"]),
        ).call()
        return result[0] / 1_000_000

    def get_positions(self, address):
        """
        Get the positions for an address
        """
        return self.opyn_controller.functions.getAccountPositions(address).call()



"""
Collateral supported by the protocol.
"""
from enum import Enum

from web3 import Web3


class Collateral(Enum):
    """Collateral supported by the protocol"""

    WETH = Web3.toChecksumAddress("0x3b3a1de07439eeb04492fa64a889ee25a130cdd3")
    USDC = Web3.toChecksumAddress("0x408c5755b5c7a0a28d851558ea3636cfc5b5b19d")

    @classmethod
    def is_supported(cls, collateral):
        """Is the sset supported"""
        try:
            cls.from_symbol(collateral)
        except ValueError:
            return False
        return True

    @classmethod
    def from_symbol(cls, symbol):
        """Returns the address of the collateral from its symbol"""
        if symbol.upper() == "WETH":
            return cls.WETH
        if symbol.upper() == "USDC":
            return cls.USDC
        raise ValueError(f"Collateral {symbol} not supported")



"""
This file contains the ActionType enum, which is used to specify the type of action to be performed by the client.
"""
from enum import Enum


class ActionType(Enum):
    """Enum to specify the type of action to be performed by the client."""

    OPEN_VAULT = 0
    MINT_SHORT_OPTION = 1
    BURN_SHORT_OPTION = 2
    DEPOSIT_LONG_OPTION = 3
    WITHDRAW_LONG_OPTION = 4
    DEPOSIT_COLLATERAL = 5
    WITHDRAW_COLLATERAL = 6
    SETTLE_VAULT = 7
    REDEEM = 8
    CALL = 9
    LIQUIDATE = 10



"""
This file contains all the constants used in the rysk_client package.
"""
DEFAULT_TIMEOUT = 10
DEFAULT_ENCODING = "utf-8"
SUBGRAPH_URL = "https://api.goldsky.com/api/public/project_clhf7zaco0n9j490ce421agn4/subgraphs/devey/0.0.2/gn"
NULL_ADDRESS = "0x0000000000000000000000000000000000000000"
SUPPORTED_LEVERAGES = [1, 1.5, 2, 3]

raw_data = {
    "localhost": {},
    "arbitrum": {
        "OpynController": "0x594bD4eC29F7900AE29549c140Ac53b5240d4019",
        "OpynAddressBook": "0xCa19F26c52b11186B4b1e76a662a14DA5149EA5a",
        "OpynOracle": "0xBA1880CFFE38DD13771CB03De896460baf7dA1E7",
        "OpynNewCalculator": "0x749a3624ad2a001F935E3319743f53Ecc7466358",
        "OpynOptionRegistry": "0x04706DE6cE851a284b569EBaE2e258225D952368",
        "priceFeed": "0xA5a095f2a2Beb2d53382293b0FfE0f520dDEC297",
        "volFeed": "0x3099900e3E9Fa62B291586f5046A09CF5b0Bccb9",
        "optionProtocol": "0x08674f64DaC31f36828B63A4468A3AC3C68Db5B2",
        "liquidityPool": "0xC10B976C671Ce9bFf0723611F01422ACbAe100A5",
        "authority": "0x0c83E447dc7f4045b8717d5321056D4e9E86dCD2",
        "portfolioValuesFeed": "0x14eF340B33bD4f64C160E3bfcD2B84D67E9b33dF",
        "optionHandler": "0xA802795269588bf33739816f76B53fD6cd099b27",
        "opynInteractions": "0x048603543a0FD41B56B831B80981Addb19C1Ea30",
        "normDist": "0xee4CfA50123109DF8DBA8CeB37d3eA94addC4A02",
        "BlackScholes": "0x2c215B6BaC6a4871C2e58669F0437853Da500020",
        "optionsCompute": "0x303956BCC420B3b74b861874d39BaD5d5eE341f0",
        "accounting": "0xd527BE017Be2C3d3d14D6bdF5C796E26bA0c5EE8",
        "uniswapV3HedgingReactor": "0x933589C46233Efa8cCDe8287E077cA6CC51Bec17",
        "perpHedgingReactor": "0xDd418b4Ec8396191D08957bD42F549e215B8e89a",
        "USDC": "0xFF970A61A04b1cA14834A43f5dE4533eBDDB5CC8",
        "WETH": "0x82aF49447D8a07e3bd95BD0d56f35241523fBab1",
    },
    "arbitrumGoerli": {
        "OpynController": "0x11a602a5F5D823c103bb8b7184e22391Aae5F4C2",
        "OpynAddressBook": "0xd6e67bF0b1Cdb34C37f31A2652812CB30746a94A",
        "OpynOracle": "0x35578F5A49E1f1Cf34ed780B46A0BdABA23D4C0b",
        "OpynNewCalculator": "0xcD270e755C2653e806e16dD3f78E16C89B7a1c9e",
        "OpynOptionRegistry": "0x48A74b742bd97545ace8B0876F5BA7ED19DF6579",
        "priceFeed": "0xDcA6c35228acb82363406CB2e7eee81B40c692aE",
        "volFeed": "0x9Fc909273C6aF5b6fFd389Fa2B44492ff88a3be6",
        "optionProtocol": "0x865Bd85b7275a33C87E8a7E31a125DD6338e6747",
        "liquidityPool": "0x2ceDe96cd46C9B751EeB868A57FEDeD060Dbe6Bf",
        "authority": "0xA524f4F9046a243c67A07dDE2D9477bf320Ed89E",
        "portfolioValuesFeed": "0xbFC1eDc5c07ada83e0244b37A784486633910cD7",
        "optionHandler": "0x8a265fa22aa5AF86fa763dC2cF04661bf06A52E6",
        "opynInteractions": "0x5e5A98E2F7c71B159B9Ed771E4138d4B2464708c",
        "normDist": "0xf63395764CAA20a54bF33fdCCdbC03b2BDc67453",
        "BlackScholes": "0x8D3ac248d2598c0C43e1D48e5ad59093F9Cb1d40",
        "optionsCompute": "0x8090b463afd758B59089905dc956300824cb2388",
        "accounting": "0xCCc6ebF870aCcD83043e8dafaeB4366A4D1a2836",
        "uniswapV3HedgingReactor": "0xE5b8F3b414a80b8C40ba438292Eb37918324d285",
        "perpHedgingReactor": "0x34f5c89fC6b053728F264cd7f9F1cdFB887AEf86",
        "USDC": "0x6775842ae82bf2f0f987b10526768ad89d79536e",
        "WETH": "0x53320bE2A35649E9B2a0f244f9E9474929d3B699",
    },
}

ADDRESSES = {
    "beyond_pricer": {
        "path": "./contracts/BeyondPricer.sol/BeyondPricer.json",
        "address": "0xc939df369C0Fc240C975A6dEEEE77d87bCFaC259",
    },
    "opyn_controller": {
        "path": "contracts/packages/opyn/core/Controller.sol/Controller.json",
        "address": "0x11a602a5F5D823c103bb8b7184e22391Aae5F4C2",
    },
    "option_exchange": {
        "path": "contracts/OptionExchange.sol/OptionExchange.json",
        "address": "0xb672fE86693bF6f3b034730f5d2C77C8844d6b45",
    },
    "option_registry": {
        "path": "contracts/OptionRegistry.sol/OptionRegistry.json",
        "address": "0x48A74b742bd97545ace8B0876F5BA7ED19DF6579",
    },
    "option_catalogue": {
        "path": "contracts/OptionCatalogue.sol/OptionCatalogue.json",
        "address": "0xde458dD32651F27A8895D4a92B7798Cdc4EbF2f0",
    },
    "usdc": {
        "path": "contracts/interfaces/I_ERC20.sol/I_ERC20.json",
        "address": "0x6775842AE82BF2F0f987b10526768Ad89d79536E",
    },
    "weth": {
        "path": "contracts/interfaces/I_ERC20.sol/I_ERC20.json",
        "address": "0x53320bE2A35649E9B2a0f244f9E9474929d3B699",
    },
}

RPC_URL = "https://arbitrum-goerli.rpc.thirdweb.com"



"""
Simple client for the rysk contracts implemented in python.
"""
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional

from rysk_client.src.subgraph import SubgraphClient
from rysk_client.web3_client import Web3Client

PRICE_DEVISOR = 1_000_000_000_000_000_000
EXPOSURE_DEVISOR = 1_000_000_000_000_000_000


def from_timestamp(date_string):
    """Parse a timestamp. 2023-05-31T08:00:00.000Z"""
    return datetime.fromtimestamp(int(date_string)).strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def to_human_format(row):
    """
    Format the row to align to the ccxt unified client.
    'ETH-16MAY23-1550-C'
    """

    month_code = row["expiration_datetime"].strftime("%b").upper()
    day = row["expiration_datetime"].strftime("%d")
    year = str(row["expiration_datetime"].year)[2:]
    strike_price = str(int(int(row["strike"]) / PRICE_DEVISOR))

    return f"ETH-{day}{month_code}{year}-{strike_price}-{'P' if row['isPut'] else 'C'}"


@dataclass
class EthCrypto:
    """Represents a crypto wallet."""

    address: Optional[str]
    private_key: Optional[str]


@dataclass
class RyskClient:
    """
    Client for the rysk contracts.
    """

    _markets: List[Dict[str, Any]]
    _tickers: List[Dict[str, Any]]

    def __init__(
        self, address: Optional[str] = None, private_key: Optional[str] = None
    ):
        self.subgraph_client = SubgraphClient()
        self.web3_client = Web3Client()
        self.verbose = True
        self._markets = []
        self._tickers = []
        self._crypto = EthCrypto(address, private_key)

    def fetch_markets(self) -> List[Dict[str, Any]]:
        """
        Fetchs the markets from the subgraph.
        """

        raw_data = self.subgraph_client.query_markets()
        data = map(self._parse_market, raw_data)

        filtered_data = filter(lambda x: x["active"], data)
        self._markets = list(filtered_data)
        return self._markets

    def _parse_market(self, raw_data: Dict[str, Any]) -> Dict[str, Any]:
        """Parses the raw data from the subgraph into a market
        {'active': True,
         'base': 'ETH',
         'baseId': 'ETH',
         'contract': True,
         'contractSize': 1.0,
         'expiry': 1685520000000,
         'expiryDatetime': '2023-05-31T08:00:00.000Z',
         'future': False,
         'id': 'ETH-31MAY23-1950-P',
         'info': {'base_currency': 'ETH',
                  'block_trade_commission': '0.0003',
                  'block_trade_min_trade_amount': '250',
                  'block_trade_tick_size': '0.0001',
                  'contract_size': '1.0',
                  'counter_currency': 'USD',
                  'creation_timestamp': '1685260860000',
                  'expiration_timestamp': '1685520000000',
                  'instrument_id': '258005',
                  'instrument_name': 'ETH-31MAY23-1950-P',
                  'instrument_type': 'reversed',
                  'is_active': True,
                  'kind': 'option',
                  'maker_commission': '0.0003',
                  'min_trade_amount': '1',
                  'option_type': 'put',
                  'price_index': 'eth_usd',
                  'quote_currency': 'ETH',
                  'rfq': False,
                  'settlement_currency': 'ETH',
                  'settlement_period': 'day',
                  'strike': '1950.0',
                  'taker_commission': '0.0003',
                  'tick_size': '0.0005'},
         'inverse': True,
         'limits': {'amount': {'max': None, 'min': 1.0},
                    'cost': {'max': None, 'min': None},
                    'leverage': {'max': None, 'min': None},
                    'price': {'max': None, 'min': 0.0005}},
         'linear': False,
         'maker': 0.0003,
         'margin': False,
         'option': True,
         'optionType': 'put',
         'precision': {'amount': 1.0, 'price': 0.0005},
         'quote': 'USD',
         'quoteId': 'USD',
         'settle': 'ETH',
         'settleId': 'ETH',
         'spot': False,
         'strike': 1950.0,
         'swap': False,
         'symbol': 'ETH/USD:ETH-230531-1950-P',
         'taker': 0.0003,
         'type': 'option'}
        """
        default = {
            "base": "ETH",
            "baseId": "ETH",
            "contract": True,
            "contractSize": 1.0,
            "spot": False,
            "swap": False,
            "future": False,
            "type": "option",
            "linear": False,
            "inverse": True,
        }
        active = all(
            [
                raw_data["isBuyable"],
                raw_data["isSellable"],
                int(raw_data["expiration"]) > datetime.now().timestamp(),
            ]
        )
        expiration_datetime = from_timestamp(raw_data["expiration"])
        raw_data.update(
            {
                "expiration_datetime": datetime.strptime(
                    expiration_datetime, "%Y-%m-%dT%H:%M:%S.%fZ"
                )
            }
        )

        default.update(
            {
                "active": active,
                "id": to_human_format(raw_data),
                "strike": int(raw_data["strike"]) / PRICE_DEVISOR,
                "optionType": "put" if raw_data["isPut"] else "call",
                "expiry": int(raw_data["expiration"]) * 1000,
                "expiryDatetime": expiration_datetime,
                "info": raw_data,
                "symbol": to_human_format(raw_data),
                "maker": 0.0003,
                "taker": 0.0003,
            }
        )
        return default

    def fetch_tickers(self) -> List[Dict[str, Any]]:
        """
        Fetchs the ticker from the beyond pricer smart contract.
            {'ask': None,
             'askVolume': 0.0,
             'average': None,
             'baseVolume': None,
             'bid': 0.03,
             'bidVolume': 2.0,
             'change': None,
             'close': None,
             'datetime': '2023-05-28T19:48:46.855Z',
             'high': None,
             'info': {'ask_iv': '0.0',
                      'best_ask_amount': '0.0',
                      'best_ask_price': '0.0',
                      'best_bid_amount': '2.0',
                      'best_bid_price': '0.03',
                      'bid_iv': '0.0',
                      'estimated_delivery_price': '1847.72',
                      'greeks': {'delta': '-0.90958',
                                 'gamma': '0.00225',
                                 'rho': '-0.12269',
                                 'theta': '-2.34938',
                                 'vega': '0.24977'},
                      'index_price': '1847.72',
                      'instrument_name': 'ETH-31MAY23-1950-P',
                      'interest_rate': '0.0',
                      'last_price': None,
                      'mark_iv': '47.17',
                      'mark_price': '0.0561',
                      'max_price': '0.0925',
                      'min_price': '0.0245',
                      'open_interest': '0.0',
                      'state': 'open',
                      'stats': {'high': None,
                                'low': None,
                                'price_change': None,
                                'volume': '0.0',
                                'volume_usd': '0.0'},
                      'timestamp': '1685303326855',
                      'underlying_index': 'SYN.ETH-31MAY23',
                      'underlying_price': '1849.1787'},
             'last': None,
             'low': None,
             'open': None,
             'percentage': None,
             'previousClose': None,
             'quoteVolume': 0.0,
             'symbol': 'ETH/USD:ETH-230531-1950-P',
             'timestamp': 1685303326855,
             'vwap': None}

        """

        if not self._markets:
            self.fetch_markets()

        tickers = map(self._fetch_ticker, self._markets)

        self._tickers = list(tickers)
        return self._tickers

    def _fetch_ticker(self, market: Dict[str, Any]) -> Dict[str, Any]:
        """
        interact with the web3 api to fetch the ticker data
        """
        ask = self.web3_client.get_options_prices(market["info"])
        bid = self.web3_client.get_options_prices(market["info"], side="sell")
        return {"ask": ask, "bid": bid, "info": market}

    def fetch_positions(self) -> List[Dict[str, Any]]:
        """
        Fetchs the positions from the opyn smart contract.
        """
        return []
